import java.util.Arrays;
import java.util.Random;

/**
   Define and demonstrate a heapsort
   algorithm on integer arrays.
*/
public class HeapSort
{
   public static void main(String[] args)
   {
      final int[] a = randomIntArray(30, 100);
      System.out.println( Arrays.toString(a) );

      sort(a);
      System.out.println( Arrays.toString(a) );
   }


   /**
      Create an array filled with random values.

      @param length  the length of the returned array
      @param n       the number of possible random values
      @return an array filled with length numbers
              between 0 and n - 1
   */
   public static int[] randomIntArray(int length, int n)
   {
      final Random generator = new Random();

      final int[] a = new int[length];
      for (int i = 0; i < a.length; ++i)
      {
         a[i] = generator.nextInt(n);
      }
      return a;
   }


   /**
      Sort an array using heapsort.

      This method will run in O(n*lg(n)) time.

      @param a  the array to sort
   */
   public static void sort(int[] a)
   {
      buildMaxHeap(a); // O(n*lg(n))

      for (int heapSize = a.length; heapSize > 1; --heapSize) // O(n*lg(n))
      {
         swap(a, 0, heapSize - 1);
         maxHeapify(a, 0, heapSize - 1); // O(lg(n))
      }
   }


   /**
      Build a max-heap out of an array if integers.

      This method will run in O(n*lg(n)) time.

      @param a  the array to turn into a max-heap
   */
   public static void buildMaxHeap(int[] a)
   {
      // Find the node in the binary tree that is
      // the last node with at least one child
      // (that is, the last node that's not a leaf).
      final int lastParent = (a.length - 2) / 2;

      // Starting from lastParent, move up the
      // binary tree heapifying every sub-tree.
      for (int i = lastParent; i >= 0; --i)
      {
         maxHeapify(a, i, a.length);  // O(lg(n))
      }
   }


   /**
      This method assumes that the binary tree at root has a
      max-heap in each of its left and right sub trees, but the
      root binary tree may not be a max-heap.

      If the root is not a max-heap, this method pushes the
      root value down into the appropriate child sub tree
      and then calls itself recursively on that child.

      This method will run in O(lg(n)) time.

      @param a     the array to heapify
      @param root  the root of the binary tree being heapifyed
      @param size  the size of the heap in the array
   */
   public static void maxHeapify(int a[], int root, int heapSize)
   {
      final int left  = 2 * root + 1; // Index of left child of root.
      final int right = 2 * root + 2; // Index of right child of root.

      int largest; // Find where the largest value is in the
                   // root node and its two child nodes.
      if (left < heapSize && a[left] > a[root])
      {
         largest = left;
      }
      else
      {
         largest = root;
      }
      if (right < heapSize && a[right] > a[largest])
      {
         largest = right;
      }

      if (largest != root) // if root is not a max-heap
      {
         swap(a, root, largest);
         maxHeapify(a, largest, heapSize);
      }
   }


   /**
      Swap two entries of an array.

      @param a  the array
      @param i  the first position to swap
      @param j  the second position to swap
   */
   private static void swap(int[] a, int i, int j)
   {
      final int temp = a[i];
      a[i] = a[j];
      a[j] = temp;
   }
}
