import java.util.Random;
import java.util.Arrays;

/**
   Define an optimized merge sort algorithm that
   switches to insertion sort on small arrays.
*/
public class QuickSort_Optimized
{
   /**
      This main() method initializes an array with random values
      and then calls the quicksort algorithm on the array.
   */
   public static void main(String[] args)
   {
      final int[] a = QuickSort.randomIntArray(32, 100);
      System.out.println( Arrays.toString(a) );

      sort(a);
      System.out.println( Arrays.toString(a) );
   }


   /**
      Sort an array using quicksort.

      Randomize the array before sorting.

      @param a  the array to sort
   */
   public static void sort(int[] a)
   {
      shuffle(a);

      sort(a, 0, a.length - 1); // sort the whole array
   }


   /**
      Sort a portion of an array using quicksort.

      Switch to insertion sort when the portion
      of the array is sufficiently small.

      @param a   the array to sort
      @param lo  the first index of the portion to be sorted
      @param hi  the last index of the portion to be sorted
   */
   public static void sort(int[] a, int lo, int hi)
   {
      if (hi - lo <= 8) // base case
      {
         insertionSort(a, lo, hi); // switch to insertion sort
         return;
      }
      final int p = partition(a, lo, hi); // Divide.
      sort(a, lo,  p);                    // Conquer left part.
      sort(a, p+1, hi);                   // Conquer right part.
   }                                      // No combine step.


   /**
      Partition a portion of an array.

      @param a   the array to partition
      @param lo  the first index of the portion to be partitioned
      @param hi  the last index of the portion to be partitioned
      @return the last index of the first partition
   */
   private static int partition(int[] a, int lo, int hi)
   {
      final int pivot = a[lo];
      int i = lo - 1;
      int j = hi + 1;
      while (i < j)
      {
         ++i; while (a[i] < pivot) { ++i; }
         --j; while (a[j] > pivot) { --j; }

         if (i < j)
         {
            swap(a, i, j);
         }
      }
      return j;
   }


   /**
      Swap two entries of an array.

      @param a  the array
      @param i  the first position to swap
      @param j  the second position to swap
   */
   private static void swap(int[] a, int i, int j)
   {
      final int temp = a[i];
      a[i] = a[j];
      a[j] = temp;
   }


   /**
      Randomize an array (shuffle the array).

      @param a  the array to shuffle
   */
   private static void shuffle(int[] a)
   {
      final Random generator = new Random();

      for (int i = a.length; i > 0; --i)
      {
         final int randomIndex = generator.nextInt(i);
         swap(a, i-1, randomIndex);
      }
   }


   /**
      Sort a portion of an array using iterative insertion sort.

      @param a   the array to sort
      @param lo  the first index of the portion to be sorted
      @param hi  the last index of the portion to be sorted
   */
   private static void insertionSort(int[] a, int lo, int hi)
   {
      for (int i = lo+1; i <= hi; ++i)
      {
         final int next = a[i];
         // Move all larger elements up.
         int j = i;
         while (j > 0 && a[j - 1] > next)
         {
            a[j] = a[j - 1];
            j--;
         }
         // Insert the element.
         a[j] = next;
      }
   }
}
